/* Functions for the so called frontend (the thing that tunes to channels)
 * $Id: fe.c,v 1.9 2011/03/19 10:01:11 simimeie Exp $
 */

#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <time.h>

#include <linux/dvb/dmx.h>
#include <linux/dvb/frontend.h>
#include <linux/dvb/version.h>

#include "getstream.h"
#include "dvbrc.h"
#include "log.h"
#include "fe.h"

static int fdf; /* Filedescriptor for the frontend */

/*
 * Dump FrontEnd Status byte as clear text returned
 * by GET_STATUS or GET_EVENT ioctl
 */
static char * fe_decode_status(fe_status_t status) {
  static char str[256]; /* Long enough for all stati combined */

  str[0] = '\0';

  if (status & FE_HAS_SIGNAL) {  strcat(str, "HAS_SIGNAL "); }
  if (status & FE_HAS_CARRIER) { strcat(str, "HAS_CARRIER "); }
  if (status & FE_HAS_VITERBI) { strcat(str, "HAS_VITERBI "); }
  if (status & FE_HAS_SYNC) {    strcat(str, "HAS_SYNC "); }
  if (status & FE_HAS_LOCK) {    strcat(str, "HAS_LOCK "); }
  if (status & FE_TIMEDOUT) {    strcat(str, "TIMEDOUT "); }
  if (status & FE_REINIT) {      strcat(str, "REINIT "); }
  
  if (str[0] == 0) { strcpy(str, "(NULL)"); }
  
  return str;
}

static unsigned char * fe_getstatusastext(void) {
  int res;
  fe_status_t festatus;
  uint16_t sigstr; /* Signal strength */
  uint16_t snr; /* Signal Noise Ratio */
  uint32_t ber; /* Bit Error Rate */
  static unsigned char sttxt[255];
  
  res = ioctl(fdf, FE_READ_STATUS, &festatus);
  if (res < 0) {
    sprintf(sttxt, "ERROR: Failed to read status from frontend.\n");
  } else {
    if (ioctl(fdf, FE_READ_SIGNAL_STRENGTH, &sigstr) < 0) sigstr = 0;
    if (ioctl(fdf, FE_READ_SNR, &snr) < 0) snr = 0;
    if (ioctl(fdf, FE_READ_BER, &ber) < 0) ber = 0;
    sprintf(sttxt,
            "Frontend status: 0x%02x %s\n"
            "Signal strength: %d   SNR: %3.0f%%   BER: %lx\n",
            festatus, fe_decode_status(festatus),
            (unsigned int)sigstr, (double)snr / 653.35, (unsigned long)ber);
  }
  return sttxt;
}

void fe_showstatus(void) {
  printf("%s", fe_getstatusastext());
}

/* Wait for the frontend to deliver an event
 * Return to callee if either TIMEDOUT or HAS_LOCK
 */
static int fe_waitforreception(void) {
  struct pollfd pfd;
  struct dvb_frontend_event event;
  int	res;
  time_t lastprint = 0;

  pfd.fd = fdf;
  pfd.events = POLLIN;

  do {
    res = poll(&pfd, 1, 1000);

    if ((console_loglevel > 0) || (file_loglevel > 0)) {
      if ((time(NULL) - lastprint) >= 2) {
        char * saveptr = NULL;
        unsigned char * stastx = fe_getstatusastext();
        char * stpt;
        stpt = strtok_r(stastx, "\n", &saveptr);
        while (stpt != NULL) {
          if (stpt[0] != '\0') {
            putlog(1, "%s", stpt);
          }
          stpt = strtok_r(NULL, "\n", &saveptr);
        }
        lastprint = time(NULL);
      }
    }
    
    if (res == 0) { continue; }
    if (res < 0) {
      perror("ERROR on poll()");
      exit(-1);
    }
    /* Free to reuse res after this */

    if (!(pfd.revents & POLLIN)) { continue; }

    res = ioctl(fdf, FE_GET_EVENT, &event);

    if (res < 0) { /* ERROR */
      /* EOVERFLOW happens if we didnt poll for events
       * lately and the event queue (8) overflowed inbetween. */
      if (errno != EOVERFLOW) {
        perror("fatal ERROR on ioctl() FE_GET_EVENT");
        exit(-1);
      } else {
        putlog(0, "Warning: %s: event queue overflowed.", __FUNCTION__);
      }
    }

    if ((res >= 0) && (event.status)) { /* No error, and an actual event */
      putlog(0, "Frontend status: 0x%02x %s", event.status, fe_decode_status(event.status));
    }

  } while (((event.status & FE_HAS_LOCK) == 0) && ((event.status & FE_TIMEDOUT) == 0));

  return event.status;
}

/* Gets called by fetune to handle dvb-s2. */
static int fe_trys2tune(uint32_t frequency, struct transponder * t) {
#if (DVB_API_VERSION < 5)
  fprintf(stderr, "ERROR: The channel you try to tune to is a DVB-S2 channel.\n");
  fprintf(stderr, "       However, the kernel headers with which this program\n");
  fprintf(stderr, "       was compiled do not support S2API.\n");
  fprintf(stderr, "       S2API is needed to support newer standards, like e.g. DVB-S2\n");
  fprintf(stderr, "       For using DVB-S2, you will need a kernel that supports\n");
  fprintf(stderr, "       the S2API (any vanilla kernel >= 2.6.28 should), and\n");
  fprintf(stderr, "       also recompile this program against the headers for such\n");
  fprintf(stderr, "       a kernel.\n");
  exit(-1);
#else
  struct dtv_property p[] = {
      { .cmd = DTV_DELIVERY_SYSTEM,   .u.data = SYS_DVBS2 },
      { .cmd = DTV_FREQUENCY,         .u.data = frequency },
      { .cmd = DTV_SYMBOL_RATE,       .u.data = t->tp.s.srate },
      { .cmd = DTV_INNER_FEC,         .u.data = FEC_AUTO },
      { .cmd = DTV_INVERSION,         .u.data = INVERSION_AUTO },
      { .cmd = DTV_PILOT,             .u.data = PILOT_AUTO },
      { .cmd = DTV_TUNE },
  };
  struct dtv_properties cmdseq = {
      .num = 7,
      .props = p
  };

  if ((ioctl(fdf, FE_SET_PROPERTY, &cmdseq)) == -1) {
    perror("FE_SET_PROPERTY failed");
    exit(-1);
  }
  
  return fe_waitforreception();
#endif
}

int fe_tune(int adapter, struct transponder * t) {

	struct dvb_frontend_parameters	feparams;
	struct lnb_s			*lnb;
	char				fename[128];
	fe_sec_tone_mode_t		tone;
	fe_sec_voltage_t		voltage;

	memset(&feparams, 0, sizeof(struct dvb_frontend_parameters));

	sprintf(fename, "/dev/dvb/adapter%d/frontend0", adapter);

	fdf = open(fename, O_RDWR);

	if (!fdf) {
		perror("Error opening dvb frontend");
		exit(-1);
	}

	if ((t->type == DVB_S) || (t->type == DVB_S2)) {
	
		/* V/H selection and 22kHz tone are currently
		 * set through old API instead, mainly because it's unclear
		 * whether they were in the first S2API version. */
		/* Select vertical (13 V) or horizontal polarity (18 V) */
		voltage = (t->tp.s.pol == 'H') ? SEC_VOLTAGE_18 : SEC_VOLTAGE_13;
	
		if (ioctl(fdf, FE_SET_VOLTAGE, voltage) < 0) {
			perror("fatal ERROR on ioctl() FE_SET_VOLTAGE");
			putlog(5, "This error can be caused by trying to use a DVB-T or DVB-C card for DVB-S");
			exit(-1);
		}
	
		lnb = t->tp.s.satellite->lnb;
	
		if (t->tp.s.freq > 2200000) { /* Frequency contains l.osc.f. */
			if (t->tp.s.freq < lnb->slof) {
				feparams.frequency = (t->tp.s.freq - lnb->lof1);
				tone = SEC_TONE_OFF;
			} else {
				feparams.frequency = (t->tp.s.freq - lnb->lof2);
				tone = SEC_TONE_ON;
			}
		} else {
			feparams.frequency = t->tp.s.freq;
			tone = SEC_TONE_OFF; /* FIXME? Why do we even set that in this case? */
		}

		if (ioctl(fdf, FE_SET_TONE, tone) < 0) {
			perror("fatal ERROR on ioctl() FE_SET_TONE");
			exit(-1);
		}

		if (t->type == DVB_S2) {
			return fe_trys2tune(feparams.frequency, t);
		}

		feparams.inversion = INVERSION_AUTO;
		feparams.u.qpsk.symbol_rate = t->tp.s.srate;
		feparams.u.qpsk.fec_inner = FEC_AUTO;

		putlog(3, "FE_SET_FRONTEND:");
		putlog(3, "\ttone = %d", tone);
		putlog(3, "\tvoltage = %d", voltage);
		putlog(3, "\tt->tp.s.freq  = %lu", t->tp.s.freq);
		putlog(3, "\tt->tp.s.srate = %lu", t->tp.s.srate);
		putlog(3, "\tlnb->lof1 = %lu", lnb->lof1);
		putlog(3, "\tlnb->lof2 = %lu", lnb->lof2);
		putlog(3, "\tlnb->slof = %lu", lnb->slof);
		putlog(3, "\tfeparams.frequency = %d", feparams.frequency);
		putlog(3, "\tfeparams.inversion = %d", feparams.inversion);
		putlog(3, "\tfeparams.u.qpsk.symbol_rate = %d", feparams.u.qpsk.symbol_rate);
		putlog(3, "\tfeparams.u.qpsk.fec_inner = %d", feparams.u.qpsk.fec_inner);

	} else if (t->type == DVB_T) {

		feparams.inversion = INVERSION_AUTO;
		feparams.frequency = t->tp.t.freq;
	
		if        (t->tp.t.bandw == 0) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_AUTO;
		} else if (t->tp.t.bandw == 6) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
		} else if (t->tp.t.bandw == 7) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
		} else if (t->tp.t.bandw == 8) {
			feparams.u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
		} else {
			putlog(0, "WARNING: Unknown bandwidth %u", t->tp.t.bandw);
		}
	
		/* FIXME make this configurable */
		feparams.u.ofdm.code_rate_HP = FEC_AUTO;
		feparams.u.ofdm.code_rate_LP = FEC_AUTO;
	
		if        (t->tp.t.modul ==   0) {
			feparams.u.ofdm.constellation = QAM_AUTO;
		} else if (t->tp.t.modul ==  16) {
			feparams.u.ofdm.constellation = QAM_16;
		} else if (t->tp.t.modul ==  32) {
			feparams.u.ofdm.constellation = QAM_32;
		} else if (t->tp.t.modul ==  64) {
			feparams.u.ofdm.constellation = QAM_64;
		} else if (t->tp.t.modul == 128) {
			feparams.u.ofdm.constellation = QAM_128;
		} else if (t->tp.t.modul == 256) {
			feparams.u.ofdm.constellation = QAM_256;
		} else {
			putlog(0, "WARNING: Unknown modulation %u", t->tp.t.modul);
		}

		if        (t->tp.t.trans == 0) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;
		} else if (t->tp.t.trans == 2) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_2K;
		} else if (t->tp.t.trans == 8) {
			feparams.u.ofdm.transmission_mode = TRANSMISSION_MODE_8K;
		} else {
			putlog(0, "WARNING: Unknown transmission mode %u", t->tp.t.trans);
		}
	
		if        (t->tp.t.guard ==  0) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
		} else if (t->tp.t.guard ==  4) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_4;
		} else if (t->tp.t.guard ==  8) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_8;
		} else if (t->tp.t.guard == 16) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_16;
		} else if (t->tp.t.guard == 32) {
			feparams.u.ofdm.guard_interval = GUARD_INTERVAL_1_32;
		} else {
			putlog(0, "WARNING: Unknown guard interval %u", t->tp.t.guard);
		}
	
		if        (t->tp.t.hier == -1) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_NONE;
		} else if (t->tp.t.hier ==  0) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_AUTO;
		} else if (t->tp.t.hier ==  1) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_1;
		} else if (t->tp.t.hier ==  2) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_2;
		} else if (t->tp.t.hier ==  4) {
			feparams.u.ofdm.hierarchy_information = HIERARCHY_4;
		} else {
			putlog(0, "WARNING: Unknown hierarchy %u", t->tp.t.hier);
		}

	} else if (t->type == DVB_C) {
		
		feparams.inversion = INVERSION_AUTO;
		feparams.frequency = t->tp.c.freq;

		if        (t->tp.c.modul ==   0) {
			feparams.u.qam.modulation = QAM_AUTO;
		} else if (t->tp.c.modul ==  16) {
			feparams.u.qam.modulation = QAM_16;
		} else if (t->tp.c.modul ==  32) {
			feparams.u.qam.modulation = QAM_32;
		} else if (t->tp.c.modul ==  64) {
			feparams.u.qam.modulation = QAM_64;
		} else if (t->tp.c.modul == 128) {
			feparams.u.qam.modulation = QAM_128;
		} else if (t->tp.c.modul == 256) {
			feparams.u.qam.modulation = QAM_256;
		} else {
			putlog(0, "WARNING: Unknown modulation %u", t->tp.t.modul);
		}

		 /* FIXME? configurable? */
		feparams.u.qam.fec_inner = FEC_AUTO;
		feparams.u.qam.symbol_rate = t->tp.c.srate;

		putlog(3, "FE_SET_FRONTEND:");
		putlog(3, "\tfeparams.frequency = %lu", feparams.frequency);
		putlog(3, "\tmodulation = %u", t->tp.c.modul);
		putlog(3, "\tsrate = %u", t->tp.c.srate);
		putlog(3, "\tfeparams.u.qam.modulation = %lu", feparams.u.qam.modulation);

	} else {
		fprintf(stderr, "ERROR: Unknown/unimplemented DVB Transponder type:\n");
		fprintf(stderr, "       I have no idea how to tune this.\n");
		exit(-1);

	}

	if (ioctl(fdf, FE_SET_FRONTEND, &feparams) < 0) {
		perror("fatal ERROR on ioctl() FE_SET_FRONTEND");
		exit(-1);
	}

	return fe_waitforreception();
}
